/*
	File:		FileDJ.java
	
	Copyright: 	 Copyright 1999-2000 Apple Computer, Inc. All rights reserved.
	
	Disclaimer:	IMPORTANT:  This Apple software is supplied to you by Apple Computer, Inc.
				("Apple") in consideration of your agreement to the following terms, and your
				use, installation, modification or redistribution of this Apple software
				constitutes acceptance of these terms.  If you do not agree with these terms,
				please do not use, install, modify or redistribute this Apple software.

				In consideration of your agreement to abide by the following terms, and subject
				to these terms, Apple grants you a personal, non-exclusive license, under Apples
				copyrights in this original Apple software (the "Apple Software"), to use,
				reproduce, modify and redistribute the Apple Software, with or without
				modifications, in source and/or binary forms; provided that if you redistribute
				the Apple Software in its entirety and without modifications, you must retain
				this notice and the following text and disclaimers in all such redistributions of
				the Apple Software.  Neither the name, trademarks, service marks or logos of
				Apple Computer, Inc. may be used to endorse or promote products derived from the
				Apple Software without specific prior written permission from Apple.  Except as
				expressly stated in this notice, no other rights or licenses, express or implied,
				are granted by Apple herein, including but not limited to any patent rights that
				may be infringed by your derivative works or by other works in which the Apple
				Software may be incorporated.

				The Apple Software is provided by Apple on an "AS IS" basis.  APPLE MAKES NO
				WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
				WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
				PURPOSE, REGARDING THE APPLE SOFTWARE OR ITS USE AND OPERATION ALONE OR IN
				COMBINATION WITH YOUR PRODUCTS.

				IN NO EVENT SHALL APPLE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR
				CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
				GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
				ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION
				OF THE APPLE SOFTWARE, HOWEVER CAUSED AND WHETHER UNDER THEORY OF CONTRACT, TORT
				(INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN IF APPLE HAS BEEN
				ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
				
	Change History (most recent first):

*/


package com.apple.jens.radio;

import com.apple.jens.mp3.*;
import java.io.*;
import java.util.StringTokenizer;


/** The FileDJ thread reads MP3 files selected by its MusicDirector
	and puts the contents into Buffers that can be streamed by Transmitters.
*/
public class FileDJ extends DJ {

	public static final String	kPropMusicLibrary  = "FileDJ.music-library",
								kPropMusicDirector = "FileDJ.music-director",
								kPropMusicFilters  = "FileDJ.music-filters";
	

	public void setStation( Station station ) {
		super.setStation(station);
		RadioProperties props = station.getProperties();
		
		MusicLibrary library = (MusicLibrary) props.getObjectProperty(kPropMusicLibrary);
		library.initialize(station);
		
		MusicDirector director = (MusicDirector) props.getObjectProperty(kPropMusicDirector);
		if( director == null )
			director = new SequentialMusicDirector();			// last-ditch default
		director.initialize(library,station);
		fIterator = director;
		
		// Parse a chain of MusicFilters that pipe through each other:
		String filters = props.getProperty(kPropMusicFilters);
		if( filters != null ) {
			StringTokenizer parser = new StringTokenizer(filters,"|,;: ");
			while( parser.hasMoreTokens() ) {
				MusicDirectorFilter filter = (MusicDirectorFilter) props.propertyValueToObject(
																			kPropMusicDirector,
																			parser.nextToken());
				if( filter != null ) {
					filter.initialize(fIterator,station);
					fIterator = filter;
				}
			}
		}
	}
	
	
	public MP3File getNextTrack( ) {
		return fIterator.getNextTrack();
	}
	
	
	protected void spin( ) throws IOException, InterruptedException {
		try{
			MP3File track = getNextTrack();
			if( track == null )
				throw new EOFException("MusicDirector finished");	// Causes thread's run loop to exit
			int bitRate = track.getBitRate();
			if( bitRate < 1000 ) throw new MP3Exception("Illegal bitrate "+bitRate);
			LOG(1,"Now playing \""+track.getTitle()+"\" by "+track.getArtist()
					+" (\""+track.getName()+"\", "+(bitRate/1000)+"kbps)");
			FileInputStream in = new FileInputStream(track);
			
			Throttle throttle = new Throttle(getName());
			
			try{
				while(true) {
					Buffer nextBuf = getNextFreeBuffer();
					LOG(2,"Loading "+nextBuf+" from "+track+"...");
					if( ! nextBuf.readFrom(in,bitRate) )		// Will block for a while (file input)
						break;
					LOG(2,"...done loading "+nextBuf);
					
					// Don't go faster than the audio bitrate or I'll overtake all the Transmitters:
					throttle.addTime( (long)nextBuf.getLength()*8000/bitRate );
				}
			}finally{
				in.close();
			}
		}catch( IOException x ) {
			Radio.WARN(1,this,"Caught an IOException, skipping to next track:");
			x.printStackTrace(System.err);
		}catch( MP3Exception x ) {
			Radio.WARN(1,this,"Caught an MP3Exception, skipping to next track:");
			x.printStackTrace(System.err);
		}
	}
	
	
	// INSTANCE DATA:
	
	
	private MusicIterator fIterator;
	
}
