/*
	File:		RadioProperties.java
	
	Copyright: 	 Copyright 1999-2000 Apple Computer, Inc. All rights reserved.
	
	Disclaimer:	IMPORTANT:  This Apple software is supplied to you by Apple Computer, Inc.
				("Apple") in consideration of your agreement to the following terms, and your
				use, installation, modification or redistribution of this Apple software
				constitutes acceptance of these terms.  If you do not agree with these terms,
				please do not use, install, modify or redistribute this Apple software.

				In consideration of your agreement to abide by the following terms, and subject
				to these terms, Apple grants you a personal, non-exclusive license, under Apples
				copyrights in this original Apple software (the "Apple Software"), to use,
				reproduce, modify and redistribute the Apple Software, with or without
				modifications, in source and/or binary forms; provided that if you redistribute
				the Apple Software in its entirety and without modifications, you must retain
				this notice and the following text and disclaimers in all such redistributions of
				the Apple Software.  Neither the name, trademarks, service marks or logos of
				Apple Computer, Inc. may be used to endorse or promote products derived from the
				Apple Software without specific prior written permission from Apple.  Except as
				expressly stated in this notice, no other rights or licenses, express or implied,
				are granted by Apple herein, including but not limited to any patent rights that
				may be infringed by your derivative works or by other works in which the Apple
				Software may be incorporated.

				The Apple Software is provided by Apple on an "AS IS" basis.  APPLE MAKES NO
				WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
				WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
				PURPOSE, REGARDING THE APPLE SOFTWARE OR ITS USE AND OPERATION ALONE OR IN
				COMBINATION WITH YOUR PRODUCTS.

				IN NO EVENT SHALL APPLE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR
				CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
				GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
				ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION
				OF THE APPLE SOFTWARE, HOWEVER CAUSED AND WHETHER UNDER THEORY OF CONTRACT, TORT
				(INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN IF APPLE HAS BEEN
				ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
				
	Change History (most recent first):

*/


package com.apple.jens.radio;

import java.io.*;
import java.util.Properties;


/** A convenience subclass of Properties that makes it easy to read properties
	from a file, and to get property values as integers or classnames.
	
*/
	
public class RadioProperties extends Properties {

	/** Default properties. Set when the app initializes. */
	public static RadioProperties sDefaults;
	

	/** Reads properties from a file. */
	public RadioProperties( File f, Properties parent ) throws IOException {
		super(parent);
		if( f.exists() ) {
			FileInputStream in = new FileInputStream(f);
			try{
				load(in);
			}finally{
				in.close();
			}
		}
	}
	
	
	/** My version of getProperty trims leading/trailing whitespace from the value. */
	public String getProperty( String name ) {
		String value = super.getProperty(name);
		if( value != null )
			value = value.trim();
		return value;
	}
	
	
	/** Looks up an integer-valued property.
		If the property is not defined, the default value is returned.
		If the property value is not a valid number, a warning is generated
		and the default value returned. */
	public int getIntProperty( String name, int deflt ) {
		String str = getProperty(name);
		if( str == null )
			return deflt;
		else {
			try{
				return Integer.parseInt(str);
			}catch( NumberFormatException x ) {
				WARN_VALUE(name,str,"is not a valid number -- will use default value "+deflt);
				return deflt;
			}
		}
	}
	
	
	/** Looks up a classname-valued property, returning a new instance of that class.
		If the property is not defined, null is returned.
		If the class does not exist or cannot be instantiated, a warning is generated
		and null returned. */
	public Object getObjectProperty( String name ) {
		return propertyValueToObject( name, getProperty(name) );
	}
	
	
	/** Instantiates an object, given the classname read from a property.
		If the instantiation fails, a warning is printed and null is returned. */
	public Object propertyValueToObject( String propName, String className ) {
		if( className != null && className.indexOf('.') < 0 )
			className = "com.apple.jens.radio." + className;		// Default package name
			
		try{
			return Class.forName(className).newInstance();
		}catch( Throwable t ) {
			WARN_VALUE(propName,className,"which could not be instantiated:");
			if(Radio.sLogLevel>=1 )
				t.printStackTrace(System.err);
			return null;
		}
	}
	
	
	/** Looks up a filename-valued property.
		If the property is not defined, null is returned.
		If the file does not exist, a warning is generated
		and null returned. */
	public File getFileProperty( String name ) {
		String value = getProperty(name);
		if( value == null )
			return null;
		File f = new File(value);
		if( ! f.exists() ) {
			WARN_VALUE(name,value,"is not an existing file/directory");
			return null;
		}
		return f;
	}
	
	
	private void WARN_VALUE( String name, String value, String message ) {
		Radio.WARN(1,"RadioProperties","Property '"+name
									+"'s value is '"+value
									+"' which "+message);
	}
	
}
